/* misc.c */

/*********************************************************************
 TCP_talk: real-time sound transmission over a LAN
    
    Copyright (C) 1995  Arnaud Louet

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include "misc.h"
#include <termios.h>
 
void exerror(char* message)
{
  perror(message);
  exit(EXIT_FAILURE);
}

int input_timeout(int filedes, unsigned int seconds)
{
  fd_set set;
  struct timeval timeout;
  
  /* init the fd set */
  FD_ZERO(&set);
  FD_SET(filedes, &set);
  
  /* init the timeval data structure */
  timeout.tv_sec = seconds;
  timeout.tv_usec = 0;
  
  /* selects returns -1 if error, 0 if timeout, 1 if input is available */
  return select(FD_SETSIZE, &set, NULL, NULL, &timeout);
}


 /* ------------------------------------------------------------ */
 /* these two functions are used to eliminate terminal buffering */ 
 /* ------------------------------------------------------------ */

  /* Use this variable to remember original terminal attributes.  */

struct termios saved_attributes;

void reset_input_mode (void)
{
  tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
 }

void disable_input_buffering (void)
{
  struct termios tattr;
  char *name;
  
  /* Make sure stdin is a terminal.  */
  if (!isatty (STDIN_FILENO))
    {
      fprintf (stderr, "Not a terminal.\n");
      exit (EXIT_FAILURE);
    }
  
  /* Save the terminal attributes so we can restore them later.  */
  tcgetattr (STDIN_FILENO, &saved_attributes);
  atexit (reset_input_mode);
  
  /* Set the funny terminal modes.  */
  tcgetattr (STDIN_FILENO, &tattr);
  tattr.c_lflag &= ~(ICANON|ECHO); /* Clear ECHO.   */
  
  /* when set to 0, read is non blocking        */
  tattr.c_cc[VMIN] = 0; /* can also be set to 1 to make read blocking */
  tattr.c_cc[VTIME] = 0;
  tcsetattr (STDIN_FILENO, TCSAFLUSH, &tattr);
  
  stdin_flush();
    
}

int get_key()
{
  int c;
  
  if ( read( STDIN_FILENO, &c, 1) == 0)
    return 0;
  return c;
}


void stdin_flush(void)
{
  if ( tcflush(STDIN_FILENO , TCIFLUSH ) == -1 )
    exerror("stdin_flush");
}


off_t filesize(int filedesc)
{
  off_t pos;
  off_t current;
  
  current = lseek(filedesc, 0, SEEK_CUR); /* Saves current file pos. */
  pos = lseek(filedesc, 0, SEEK_END);
  lseek (filedesc, current, SEEK_SET); /* restores current file pos. */
  
  return pos;
}

void 
blk_read( int fd, void *buf,size_t count)
     {
	size_t nleft;
	size_t nread;
	char *ptr;
	
	ptr = buf;			       /*can't do pointer arithmetic on void */
	nleft = count;
	while( nleft > 0 )
	  {
	     if (( nread = read(fd, ptr, nleft)) < 0 )
	       exerror("read in blk_read");
	     else if (nread == 0)
	       {
		  fprintf(stderr,"Unexpected EOF in blk_read\n");		       /* EOF */
		  exit(EXIT_FAILURE);
	       }
	     nleft -= nread;
	     ptr += nread;
	  }
	
     }


void
blk_write( int fd, void *buf, size_t count)
     {
	size_t nleft;
	size_t nwritten;
	const char *ptr;
	
	ptr = buf;
	nleft = count;
	
	while( nleft > 0 )
	  {
	     if (( nwritten = write(fd, ptr, nleft))  <= 0 )
	       exerror("read in blk_write");
	     nleft -= nwritten;
	     ptr += nwritten;
	  }
     }

