
/*
**  jazz - a midi sequencer for Linux
**
**  Copyright (C) 1995-1996 Andreas Voss (andreas@avix.rhein-neckar.de)
**
**  Copyright (C) 1995-1996 Per Sigmond (Per.Sigmond@hia.no)
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef wx_wxh
#include <wx.h>
#endif

#include "harmony.h"
#include "harmonyp.h"
#include "player.h"
#include "trackwin.h"
#include "pianowin.h"
#include "guitar.h"
#include "song.h"
#include "filter.h"
#include "util.h"
#include "jazz.h"
#include <iostream.h>

#ifdef wx_xt
#define wxbMessageBox wxMessageBox
#endif

tHBInterface *the_harmony_browser = 0;

// ------------------------ HBPlayer ------------------------

class HBPlayer : public wxTimer
{
    friend class HBCanvas;
  public:
    HBPlayer();
    void StartPlay(const HBContext &);
    void Paste(tEventArray &);
    void StopPlay();
    void SettingsDialog(wxFrame *parent);
    int  IsPlaying() const { return playing; }
    const HBContext &Context()	{ return context; }

    virtual void Notify();

    int  GetChordKeys(int *out, const HBContext &);
    int  GetMeldyKeys(int *out, const HBContext &);
    int  GetBassKey(const HBContext &);

  private:
    HBContext context;
    static int bass_channel, bass_veloc;
    static int chord_channel, chord_veloc;
    static int meldy_channel, meldy_veloc;
    static Bool bass_enabled, chord_enabled, meldy_enabled;
    static int bass_pitch, chord_pitch, meldy_pitch;
    static int meldy_speed;

    int bass_key, chord_keys[12], n_chord_keys;
    int meldy_keys[12], n_meldy_keys, meldy_index;
    int note_length;
    int playing;
};

int HBPlayer::bass_enabled  = 1;
int HBPlayer::bass_channel  = 1;
int HBPlayer::bass_veloc    = 90;
int HBPlayer::bass_pitch    = 40;

int HBPlayer::chord_enabled = 1;
int HBPlayer::chord_channel = 2;
int HBPlayer::chord_veloc   = 90;
int HBPlayer::chord_pitch   = 60;

int HBPlayer::meldy_enabled = 0;
int HBPlayer::meldy_channel = 3;
int HBPlayer::meldy_veloc   = 90;
int HBPlayer::meldy_pitch   = 70;
int HBPlayer::meldy_speed   = 100;


HBPlayer::HBPlayer()
{
  playing = 0;
  bass_key = n_chord_keys = n_meldy_keys = 0;
  note_length = 60;
  meldy_index = 0;
}


int HBPlayer::GetChordKeys(int *out, const HBContext &context)
{
  // build chord keys
  HBChord chord = context.Chord();
  int key = chord.Iter(chord_pitch - 1);
  int n   = chord.Count();
  for (int i = 0; i < n; i++)
  {
    out[i] = key;
    key = chord.Iter(key);
  }
  return n;
}


int HBPlayer::GetBassKey(const HBContext &context)
{
  // build bass note
  int key = context.ChordKey() + bass_pitch - bass_pitch % 12;
  if (key < bass_pitch)
    key += 12;
  return key;
}


int HBPlayer::GetMeldyKeys(int *out, const HBContext &context)
{
  // build melody keys
  HBChord scale = context.Scale();
  int n = scale.Count();

  int key = scale.Iter(meldy_pitch);
  int j = 0;
  for (int i = 0; i < n; i++)
  {
    out[j++] = key;
    key = scale.Iter(key);
  }
  return n;
}

void HBPlayer::Paste(tEventArray &arr)
{
  if (bass_enabled)
  {
    tKeyOn e(0, bass_channel - 1, bass_key, bass_veloc, note_length);
    arr.Put(e.Copy());
  }

  if (chord_enabled)
  {
    for (int i = 0; i < n_chord_keys; i++)
    {
      tKeyOn e(0, chord_channel - 1, chord_keys[i], chord_veloc, note_length);
      arr.Put(e.Copy());
    }
  }
}


void HBPlayer::StartPlay(const HBContext &ct)
{
  int i;

  if (playing)
    StopPlay();
  playing = 1;
  context = ct;

  bass_key = GetBassKey(context);
  n_chord_keys = GetChordKeys(chord_keys, context);
  n_meldy_keys = GetMeldyKeys(meldy_keys, context);

  // Generate KeyOn's
  if (bass_enabled)
  {
    tKeyOn e(0, bass_channel - 1, bass_key, bass_veloc);
    Midi->OutNow(&e);
  }

  if (chord_enabled)
  {
    for (i = 0; i < n_chord_keys; i++)
    {
      tKeyOn e(0, chord_channel - 1, chord_keys[i], chord_veloc);
      Midi->OutNow(&e);
    }
  }

  Notify();
  Start(60000L / 4 / meldy_speed);
}


void HBPlayer::Notify()
{
  if (meldy_enabled)
  {
    tKeyOff of(0, meldy_channel - 1, meldy_keys[meldy_index]);
    Midi->OutNow(&of);
    meldy_index = (meldy_index + 1) % n_meldy_keys;
    tKeyOn on(0, meldy_channel - 1, meldy_keys[meldy_index], meldy_veloc);
    Midi->OutNow(&on);
  }
}


void HBPlayer::StopPlay()
{
  if (!playing) 
    return;
  Stop();
  playing = 0;

  int i;
  // Generate KeyOff's
  if (bass_enabled)
  {
    tKeyOff e(0, bass_channel - 1, bass_key);
    Midi->OutNow(&e);
  }

  if (chord_enabled)
  {
    for (i = 0; i < n_chord_keys; i++)
    {
      tKeyOff e(0, chord_channel - 1, chord_keys[i]);
      Midi->OutNow(&e);
    }
  }

  if (meldy_enabled)
  {
    for (i = 0; i < n_meldy_keys; i++)
    {
      tKeyOff of(0, meldy_channel - 1, meldy_keys[i]);
      Midi->OutNow(&of);
    }
  }
}

class tHBPlayerForm : public wxForm
{
  public:
	tHBPlayerForm() : wxForm( USED_WXFORM_BUTTONS ) {}
	void OnHelp()
	{      
	  HelpInstance->LoadFile( HelpPathList.FindValidPath("jazz") );
       	  HelpInstance->KeywordSearch("Harmony browser");
	}
};


void HBPlayer::SettingsDialog(wxFrame *parent)
{
  wxDialogBox *panel = new wxDialogBox(parent, "MIDI settings", FALSE);
  tHBPlayerForm      *form  = new tHBPlayerForm;

  form->Add(wxMakeFormMessage("Note Length for paste into piano window"));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Length", &note_length, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(10.0, 120.0), 0)));
  form->Add(wxMakeFormNewLine());

  panel->SetLabelPosition(wxHORIZONTAL);

  form->Add(wxMakeFormBool("Bass enable", &bass_enabled));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Channel", &bass_channel, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(1.0, 16.0), 0)));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Velocity", &bass_veloc, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(1.0, 127.0), 0)));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Pitch", &bass_pitch, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(30.0, 99.0), 0)));
  form->Add(wxMakeFormNewLine());

  form->Add(wxMakeFormBool("Chord enable", &chord_enabled));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Channel", &chord_channel, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(1.0, 16.0), 0)));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Velocity", &chord_veloc, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(1.0, 127.0), 0)));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Pitch", &chord_pitch, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(30.0, 99.0), 0)));
  form->Add(wxMakeFormNewLine());

  form->Add(wxMakeFormBool("Scale enable", &meldy_enabled));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Channel", &meldy_channel, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(1.0, 16.0), 0)));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Velocity", &meldy_veloc, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(1.0, 127.0), 0)));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Pitch", &meldy_pitch, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(30.0, 99.0), 0)));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormShort("Speed", &meldy_speed, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(60.0, 180.0), 0)));
  form->Add(wxMakeFormNewLine());

  form->AssociatePanel(panel);
  panel->Fit();

  panel->Show(TRUE);
}

// =============================================================
// HBCanvas
// =============================================================

class HBCanvas : public wxCanvas
{
    friend class HBSettingsDlg;
    friend class HBFrame;
    friend class HBMatchMarkers;
  public:
    HBCanvas(wxFrame *parent, int x, int y, int w, int h);
    virtual ~HBCanvas();
    virtual void OnPaint(void);
    void DrawMarkers(const HBContext &c);
    void ClearSeq();
    virtual void OnEvent(wxMouseEvent &e);
    void SettingsDialog(wxFrame *parent);

    int SeqDefined()	{ return n_seq > 0; }
    int GetChordKeys(int *out, int step, int n_steps);
    int GetSelectedChord(int *out);
    int GetSelectedScale(int *out);
    int GetBassKeys(int *out, int step, int n_steps);


    HBPlayer player;
    enum {SEQMAX = 24};

  protected:
    static const int ScFa;
    void ChordRect(tRect &r, const HBContext &ct);
    void DrawChord(const HBContext &ct);
    void UnDrawChord(const HBContext &ct);
    Bool Find(float x, float y, HBContext &out);

  private:

    float xchord, ychord, wchord, hchord;
    float ofs;
    wxFrame *parent;

    HBContext *seq[SEQMAX];
    int n_seq;

    HBContext mouse_context;

    static Bool haunschild_layout;
    static Bool mark_same;
    static Bool mark_3_common;
    static Bool mark_1_semi;
    static Bool mark_tritone;
    static Bool mark_piano;
};

Bool HBCanvas::haunschild_layout = 0;
Bool HBCanvas::mark_same = 0;
Bool HBCanvas::mark_piano = 0;
Bool HBCanvas::mark_3_common = 0;
Bool HBCanvas::mark_1_semi = 0;
Bool HBCanvas::mark_tritone = 0;

const int HBCanvas::ScFa = 50;


int HBCanvas::GetChordKeys(int *out, int step, int n_steps)
{
  if (n_seq == 0)
    return 0;
  int i = step * n_seq / n_steps;
  return player.GetChordKeys(out, *seq[i]);
}

int HBCanvas::GetSelectedChord(int *out)
{
  return player.GetChordKeys(out, mouse_context);
}

int HBCanvas::GetSelectedScale(int *out)
{
  return player.GetMeldyKeys(out, mouse_context);
}


int HBCanvas::GetBassKeys(int *out, int step, int n_steps)
{
  if (n_seq == 0)
    return 0;
  int i = step * n_seq / n_steps;
  out[0] = player.GetBassKey(*seq[i]);
  return 1;
}


HBCanvas::HBCanvas(wxFrame *p, int x, int y, int w, int h)
  : wxCanvas(p, x, y, w, h, 0)
{
  parent = p;
  n_seq  = 0;
  //seq    = new HBContext [SEQMAX];
  for (int i = 0; i < SEQMAX; i++)
    seq[i] = new HBContext();

  wxDC *dc = GetDC();
  dc->SetFont(wxSMALL_FONT);
  float tw, th;
  dc->GetTextExtent("xD#j75+9-x", &tw, &th);

  wchord = 1.2 * tw;
  hchord = 2.5 * th;
  xchord = 50;
  ychord = 4 * hchord;
  ofs    = th/4;

  SetScrollbars(0, 20, 0, (int)(ychord + hchord * (int)nScaleTypes * 12.5)/20 + 1, 8, 8);
}

HBCanvas::~HBCanvas()
{
  if (player.IsPlaying())
    player.StopPlay();
  for (int i = 0; i < SEQMAX; i++)
    delete seq[i];
}


void HBCanvas::ChordRect(tRect &r, const HBContext &ct)
{
  if (ct.SeqNr())
  {
    r.x = xchord + (ct.SeqNr() - 1) % 7 * wchord;
    r.y = hchord * ((ct.SeqNr() -1) / 7 + 0.5);
  }
  else if (!haunschild_layout)
  {
    r.x = xchord + ct.ChordNr() * wchord;
    r.y = ychord + (int)ct.ScaleType() * hchord * 13;
    r.y += ct.ScaleNr() * hchord;
  }
  else
  {
    r.x = xchord + (5 * ct.ChordNr() % 7 + 5 * ct.ScaleNr() % 12) % 7 * wchord;
    r.y = ychord + ((int)ct.ScaleType()) * hchord * 13;
    r.y += (5 * ct.ScaleNr() % 12) * hchord;
  }

  r.x += ofs;
  r.y -= ofs;
  r.w =  wchord - 2 * ofs;
  r.h =  hchord - 2 * ofs;
}


void HBCanvas::DrawChord(const HBContext &ct)
{
  // draw surrounding box
  tRect r;
  ChordRect(r, ct);
  wxDC *dc = GetDC();
  dc->DrawRectangle(r.x, r.y, r.w, r.h);

  float w, h;
  const char *name = ct.ChordName();
  dc->GetTextExtent(name, &w, &h);
  dc->DrawText((char *)name, r.x + (r.w - w)/2, r.y + (r.h - h)/2);
}


void HBCanvas::UnDrawChord(const HBContext &ct)
{
  // draw surrounding box
  tRect r;
  ChordRect(r, ct);
  wxDC *dc = GetDC();
  dc->SetPen(wxWHITE_PEN);
  dc->DrawRectangle(r.x, r.y, r.w, r.h);
  dc->SetPen(wxBLACK_PEN);
}


void HBCanvas::OnPaint()
{
  wxDC *dc = GetDC();
  dc->Clear();
  dc->DrawText("Seq", 5, 5);

  HBContextIterator iter;
  iter.SetSequence(seq, n_seq);
  while (iter())
  {
    const HBContext &ct = iter.Context();
    DrawChord(ct);
    if (ct.ChordNr() == 0 && ct.SeqNr() == 0)
    {
      tRect r;
      ChordRect(r, ct);
      dc->DrawText((char *)ct.ScaleName(), 5, r.y);
    }
  }
  DrawMarkers(mouse_context);

  if (!haunschild_layout)
  {
    for (int i = 0; i < nScaleTypes; i++)
    {
      for (int j = 0; j < 7; j++)
      {
	HBContext ct(0, j, (tScaleType)i);
	tRect r;
	ChordRect(r, ct);
	r.y -= hchord;
	float w, h;
	const char *name = ct.ChordNrName();
	dc->GetTextExtent(name, &w, &h);
	dc->DrawText((char *)name, r.x + (r.w - w)/2, r.y + (r.h - h)/2);
      }
    }
  }
}


class HBMatchMarkers : public HBMatch
{
  public:
    HBMatchMarkers(const HBContext &ct, HBCanvas *cv);
    virtual bool operator()(const HBContext &);
  private:
    HBCanvas  *cnvs;
    HBContext context;
    HBChord   chord;
    HBChord   scale;
    int       n_chord;
    int       chord_key;

    HBChord   tritone;
    HBChord   piano;
};


HBMatchMarkers::HBMatchMarkers(const HBContext &ct, HBCanvas *cv)
  : context(ct)
{
  cnvs      = cv;
  chord     = context.Chord();
  n_chord   = chord.Count();
  chord_key = context.ChordKey();
  scale     = context.Scale();

  // if there is a chord already, use it
  if (scale.Contains(chord_key + 6))
  {
    int key = chord_key + 6;
    for (int i = 0; i < 4; i++)
    {
      tritone += key;
      key = scale.Iter(key);
      key = scale.Iter(key);
    }
  }
  else
  {
    // simple transpose
    tritone = chord; 
    tritone.Rotate(6);
  }

  if (cnvs->mark_piano)
  {
    tEventArray &buf = TrackWin->GetPianoWin()->PasteBuffer;
    for (int i = 0; i < buf.nEvents; i++)
    {
      tKeyOn *on = buf.Events[i]->IsKeyOn();
      if (on)
        piano += on->Key;
    }
  }
}


bool HBMatchMarkers::operator()(const HBContext &o_context)
{
  HBChord o_chord = o_context.Chord();

  if (cnvs->mark_piano && o_chord.Contains(piano))
    return true;

  if (cnvs->mark_same && o_chord == chord)
    return true;

  if (cnvs->mark_tritone && o_chord == tritone)
    return true;

  HBChord common = (chord & o_chord);
  int n_common = common.Count();

  if (cnvs->mark_3_common && n_common == 3)
    return true;

  if (cnvs->mark_1_semi && n_common == n_chord - 1)
  {
    HBChord delta = chord ^ o_chord;
    int key = delta.Iter(0);
    if (delta.Contains(key + 1) || delta.Contains(key - 1))
      return true;
  }

  return false;
}


void HBCanvas::DrawMarkers(const HBContext &ct)
{
  tRect r;
  wxDC *dc = GetDC();
  dc->SetLogicalFunction(wxXOR);
  dc->SetBrush(wxTRANSPARENT_BRUSH);
  HBMatchMarkers match(ct, this);
  HBContextIterator iter(match);
  iter.SetSequence(seq, n_seq);
  while (iter())
  {
    ChordRect(r, iter.Context());
    r.x += 3;
    r.y += 3;
    r.w -= 6;
    r.h -= 6;
    dc->DrawRectangle(r.x, r.y, r.w, r.h);
  }

  // invert actual chord
  dc->SetBrush(wxBLACK_BRUSH);
  ChordRect(r, ct);
  dc->DrawRectangle(r.x, r.y, r.w, r.h);
  if (ct.SeqNr() > 0)
  {
    HBContext c(ct);
    c.SetSeqNr(0);
    ChordRect(r, c);
    dc->DrawRectangle(r.x, r.y, r.w, r.h);
  }
  dc->SetLogicalFunction(wxCOPY);
  dc->SetBrush(wxWHITE_BRUSH);
}


Bool HBCanvas::Find(float x, float y, HBContext &out)
{
  HBContextIterator iter;
  iter.SetSequence(seq, n_seq);
  while (iter())
  {
    tRect r;
    ChordRect(r, iter.Context());
    if (r.IsInside(x, y))
    {
      out = iter.Context();
      return TRUE;
    }
  }
  return FALSE;
}


void HBCanvas::ClearSeq()
{
  n_seq = 0;
  mouse_context.SetSeqNr(0);
  OnPaint();
}


void HBCanvas::OnEvent(wxMouseEvent &e)
{
  HBContext context;
  float x, y;
  e.Position(&x, &y);
  if (Find(x, y, context))
  {
    if (e.ButtonDown())
    {
      player.StartPlay(context);
      if (e.LeftDown() && e.ShiftDown() || e.MiddleDown())
      {
	if (context.SeqNr())
	{
	  // remove a chord
	  if (context.SeqNr() == n_seq)
	  {
	    // remove markers first
	    if (mouse_context.SeqNr() == n_seq)
	    {
	      DrawMarkers(mouse_context);
	      mouse_context.SetSeqNr(0);
	      DrawMarkers(mouse_context);
	    }
	    -- n_seq;
	    UnDrawChord(context);
	    context.SetSeqNr(0);
	  }
	}
	else if (n_seq < SEQMAX)
	{
	  // add a chord
	  context.SetSeqNr(n_seq + 1);
	  *seq[n_seq ++] = context;
	  DrawMarkers(mouse_context);
	  DrawChord(context);
	  DrawMarkers(mouse_context);
	}
      }
    }
    else if (e.Dragging() && player.IsPlaying() && context != player.Context())
    {
      player.StopPlay();
      player.StartPlay(context);
    }

    if ((e.LeftDown() || e.MiddleDown()) ) // && context != mouse_context)
    {
      DrawMarkers(mouse_context);
      mouse_context = context;
      //mouse_context.SetSeqNr(0);
      DrawMarkers(mouse_context);

      // paste to PianoWin buffer
      if (!mark_piano)
      {
	tEventArray &buf = TrackWin->GetPianoWin()->PasteBuffer;
	buf.Clear();
	player.Paste(buf);
	TrackWin->GetPianoWin()->Redraw();
      }

      // Show in GuitarWin
      tGuitarWin *guitar = TrackWin->GetPianoWin()->GetGuitarWin();
      if (guitar)
      {
        guitar->ShowPitch(0);	// remove actual pianowin/mouse position
        guitar->Redraw();
      }
    }
  }
  if (e.ButtonUp() && player.IsPlaying())
    player.StopPlay();
}


class HBSettingsForm : public wxForm
{
  public:
    HBSettingsForm(HBCanvas *c) 
	: wxForm( USED_WXFORM_BUTTONS )
    { cnvs = c; }
    virtual void OnOk() { cnvs->OnPaint(); wxForm::OnOk(); }
    virtual void OnHelp();
  private:
    HBCanvas *cnvs;
};

void HBSettingsForm::OnHelp()
{
        HelpInstance->LoadFile( HelpPathList.FindValidPath("jazz") );
        HelpInstance->KeywordSearch("Harmony browser");
}


void HBCanvas::SettingsDialog(wxFrame *parent)
{
  wxDialogBox *panel = new wxDialogBox(parent, "Chord settings", FALSE);
  wxForm      *form  = new HBSettingsForm(this);

  panel->SetLabelPosition(wxHORIZONTAL);

  form->Add(wxMakeFormBool("Haunschild layout", &haunschild_layout));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormBool("same notes", &mark_same));
  form->Add(wxMakeFormBool("1 semitone", &mark_1_semi));
  form->Add(wxMakeFormBool("3 common notes", &mark_3_common));
  form->Add(wxMakeFormBool("tritone", &mark_tritone));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormMessage("enabled = show contents, disabled = paste actual chord"));
  form->Add(wxMakeFormNewLine());
  form->Add(wxMakeFormBool("show pianowin buffer", &mark_piano));
  form->Add(wxMakeFormNewLine());

  form->AssociatePanel(panel);
  panel->Fit();
  panel->Show(TRUE);
}

// ---------------------------------------------------------
// HBContextDlg
// ---------------------------------------------------------

struct tNamedChord
{
  const char *name;
  long bits;
};

const int n_chord_names = 12;
const int n_scale_names = 47;


tNamedChord chord_names[n_chord_names] = {
  { "j7",	0x891},
  { "m7",	0x489},
  { "7",	0x491},
  { "m75-",	0x449},
  { "mj7",	0x889},
  { "j75+",	0x911},
  { "0",	0x249},
  { "sus4",	0xa1},
  { "7sus4",	0x4a1},
  { "j7sus4",	0x8a1},
  { "alt (79+13-)",	0x519},
  { "75-",	0x451},
};

tNamedChord scale_names[n_scale_names] = {
  { "***** major scales *****",	0x0},
  { "major penta",	0x295},
  { "ionic",	0xab5},
  { "lydic",	0xad5},
  { "har III (ion 5+)",	0xb35},
  { "har VI (lyd 9+)",	0xad9},
  { "mel III (lyd 5+)",	0xb55},
  { "augmented",	0x333},
  { "ionic 13-",	0x9b5},
  { "***** minor scales *****",	0x0},
  { "minor penta",	0x4a9},
  { "aeolic",	0x5ad},
  { "doric",	0x6ad},
  { "mel II (doric 9-)",	0x6ab},
  { "phrygic",	0x5ab},
  { "jap. penta",	0x4a3},
  { "har IV (dor 11+)",	0x6cd},
  { "harmonic minor",	0x9ad},
  { "melodic minor",	0xaad},
  { "gipsy",	0x9cd},
  { "melodic 11+",	0xacd},
  { "***** dominant scales *****",	0x0},
  { "major penta",	0x295},
  { "ind. penta",	0x4b1},
  { "mixolyd",	0x6b5},
  { "har V (har dominant)",	0x5b3},
  { "mel IV (mixo 11+)",	0x6d5},
  { "mixo 11+9-",	0x6d3},
  { "mel V (mixo 13-)",	0x5b5},
  { "mixo 9-",	0x6b3},
  { "full 1",	0x555},
  { "full 2",	0x555},
  { "har alt",	0x59b},
  { "mel VII (altered)",	0x55b},
  { "semi/full",	0x6db},
  { "***** semi dimin *****",	0x0},
  { "locrian",	0x56b},
  { "mel VI (locrian 9)",	0x56d},
  { "har II (locrian 13)",	0x66b},
  { "doric 5-",	0x66d},
  { "***** dimin *****",	0x0},
  { "har VII (har dim)",	0x35b},
  { "full/semi",	0xa6d},
  { "locrian dim",	0x36b},
  { "***** blues scales *****",	0x0},
  { "minor penta 5-",	0x4e9},
  { "blues scale",	0x4f9},
};


class HBContextDlg : public wxDialogBox
{
  public:
    HBContextDlg(HBCanvas *c, wxFrame *parent, HBContext *pcontext);
    ~HBContextDlg();
    static void OkButton(wxButton &but, wxCommandEvent& event);
    static void CancelButton(wxButton &but, wxCommandEvent& event);
    static void PlayButton(wxButton &but, wxCommandEvent& event);
    static void HelpButton(wxButton &but, wxCommandEvent& event);
    static void ChordCheck(wxItem &item, wxCommandEvent& event);
    static void ScaleCheck(wxItem &item, wxCommandEvent& event);
    static void ChordList(wxItem &item, wxCommandEvent& event);
    static void ScaleList(wxItem &item, wxCommandEvent& event);
    void OnOkButton();
    void OnCancelButton();
    void OnPlayButton();
    void OnChordCheck();
    void OnScaleCheck();
    void OnChordList();
    void OnScaleList();

    void OnHelp();

    void ShowValues();

  private:

    HBCanvas   *cnvs;
    wxCheckBox *chord_chk[12];
    wxCheckBox *scale_chk[12];
    wxListBox  *chord_lst;
    wxListBox  *scale_lst;
    wxMessage  *chord_msg;

    wxButton   *ok_but;
    wxButton   *cancel_but;
    wxButton   *play_but;
    wxButton   *help_but;

    HBChord    chord;
    HBChord    scale;
    int        chord_key;
    int        scale_key;
    int	       ChordKey(int i = 0) const { return (chord_key + i) % 12; }
    int	       ScaleKey(int i = 0) const { return (scale_key + i) % 12; }
    HBContext  *pcontext;

    HBPlayer   player;
    void       RestartPlayer();
};


HBContextDlg::HBContextDlg(HBCanvas *c, wxFrame *parent, HBContext *pct)
  : wxDialogBox(parent, "Edit chord/scale", FALSE)
{ 
  int i;

  cnvs = c; 
  pcontext = pct;
  chord = pcontext->Chord();
  scale = pcontext->Scale();
  chord_key = pcontext->ChordKey();
  scale_key = pcontext->ScaleKey();

  // buttons
  ok_but = new wxButton(this, (wxFunction)OkButton, "Ok") ;
  cancel_but = new wxButton(this, (wxFunction)CancelButton, "Cancel") ;
  play_but = new wxButton(this, (wxFunction)PlayButton, "Play") ;
  help_but = new wxButton(this, (wxFunction)HelpButton, "Help" );
  NewLine();

  // top messages
  (void)new wxMessage(this, "Chord: ");
  chord_msg = new wxMessage(this, "Am75-13-sus4");
  NewLine();

  // chord/scale keys 
#ifdef wx_xview
  int w = 45;
  int h = 30;
  int y = 70;
#endif

#ifdef wx_motif
  int w = 45;
  int h = 40;
  int y = 80;
#endif

#ifdef wx_xt
  int w = 45;
  int h = 40;
  int y = 80;
#endif

#ifdef wx_msw
  int w = 45;
  int h = 40;
  int y = 80;
#endif

  const char *notename[12] = { "1", 0, "9", 0, "3", "11", 0, "5", 0, "13", 0, "7" };
  for (i = 0; i < 12; i++)
  {
    int x = w * i + 10;
    chord_chk[i] = new wxCheckBox(this, (wxFunction)ChordCheck, " ", x, y+1*h);
    scale_chk[i] = new wxCheckBox(this, (wxFunction)ScaleCheck, " ", x, y+2*h);
    if (notename[i])
      (void) new wxMessage(this, (char *)notename[i], x, y+3*h);
    (void) new wxMessage(this, (char *)HBChord::ScaleName(i + chord_key), x, y+0*h);
  }
  y += 4*h;

  // list boxes                                                               x  y    w    h
  SetLabelPosition(wxVERTICAL);
  char **cnames = new char * [n_chord_names];
  for (i = 0; i < n_chord_names; i++)
    cnames[i] = (char *)chord_names[i].name;
  chord_lst = new wxListBox(this, (wxFunction)ChordList, "Chords", wxSINGLE, 10, y, 100, 200, n_chord_names, cnames, wxNEEDED_SB);
  delete [] cnames;

  char **snames = new char * [n_scale_names];
  for (i = 0; i < n_scale_names; i++)
    snames[i] = (char *)scale_names[i].name;
  scale_lst = new wxListBox(this, (wxFunction)ScaleList, "Scales", wxSINGLE, 200, y, 300, 200, n_scale_names, snames, wxNEEDED_SB);
  delete [] snames;

  // thats it
  Fit();
  Show(TRUE);
  ShowValues();
}

HBContextDlg::~HBContextDlg()
{
  if (player.IsPlaying())
    player.StopPlay();
}

void HBContextDlg::ShowValues()
{
  char buf[30];
  chord.Name(buf, ChordKey(0));
  chord_msg->SetLabel(buf);
  for (int i = 0; i < 12; i++)
  {
    chord_chk[i]->SetValue(0 != chord.Contains(ChordKey(i)));
    scale_chk[i]->SetValue(0 != scale.Contains(ScaleKey(i)));
  }
}

void HBContextDlg::OnOkButton()
{
  chord.Clear();
  scale.Clear();
  for (int i = 0; i < 12; i++)
  {
    if (chord_chk[i]->GetValue())
      chord += ChordKey(i);
    if (scale_chk[i]->GetValue())
      scale += ScaleKey(i);
  }
  *pcontext->PChord() = chord;
  *pcontext->PScale() = scale;
  cnvs->OnPaint();
  delete this;
}

void HBContextDlg::OnCancelButton()
{
  delete this;
}

void HBContextDlg::OnPlayButton()
{
  if (player.IsPlaying())
  {
    play_but->SetLabel("play");
    player.StopPlay();
  }
  else
  {
    HBContext ct(*pcontext);
    *ct.PChord() = chord;
    *ct.PScale() = scale;
    player.StartPlay(ct);
    play_but->SetLabel("stop");
  }
}

void HBContextDlg::OnHelp()
{
        HelpInstance->LoadFile( HelpPathList.FindValidPath("jazz") );
        HelpInstance->KeywordSearch("Edit chord");
}


void HBContextDlg::RestartPlayer()
{
  if (player.IsPlaying())
  {
    player.StopPlay();
    HBContext ct(*pcontext);
    *ct.PChord() = chord;
    *ct.PScale() = scale;
    player.StartPlay(ct);
  }
}

void HBContextDlg::OnChordCheck()
{
  char buf[30];
  chord.Clear();
  for (int i = 0; i < 12; i++)
  {
    if (chord_chk[i]->GetValue())
      chord += ChordKey(i);
  }
  chord.Name(buf, ChordKey());
  chord_msg->SetLabel(buf);
  RestartPlayer();
}


void HBContextDlg::OnScaleCheck()
{
  scale.Clear();
  for (int i = 0; i < 12; i++)
  {
    if (scale_chk[i]->GetValue())
      scale += ScaleKey(i);
  }
  RestartPlayer();
}

void HBContextDlg::OnChordList()
{
  int i = chord_lst->GetSelection();
  if (i >= 0)
  {
    HBChord c(chord_names[i].bits);
    c.Rotate(ChordKey());
    chord = c;
    ShowValues();
    RestartPlayer();
  }
}


void HBContextDlg::OnScaleList()
{
  int i = scale_lst->GetSelection();
  if (i >= 0)
  {
    HBChord s(scale_names[i].bits);
    s.Rotate(ScaleKey());
    scale = s;
    ShowValues();
    RestartPlayer();
  }
}


void HBContextDlg::OkButton(wxButton &but, wxCommandEvent& event)
{
  ((HBContextDlg *)but.GetParent())->OnOkButton();
}

void HBContextDlg::HelpButton(wxButton &but, wxCommandEvent& event)
{
  ((HBContextDlg *)but.GetParent())->OnHelp();
}

void HBContextDlg::CancelButton(wxButton &but, wxCommandEvent& event)
{
  ((HBContextDlg *)but.GetParent())->OnCancelButton();
}

void HBContextDlg::PlayButton(wxButton &but, wxCommandEvent& event)
{
  ((HBContextDlg *)but.GetParent())->OnPlayButton();
}

void HBContextDlg::ChordCheck(wxItem &itm, wxCommandEvent& event)
{
  ((HBContextDlg *)itm.GetParent())->OnChordCheck();
}

void HBContextDlg::ScaleCheck(wxItem &itm, wxCommandEvent& event)
{
  ((HBContextDlg *)itm.GetParent())->OnScaleCheck();
}

void HBContextDlg::ChordList(wxItem &itm, wxCommandEvent& event)
{
  ((HBContextDlg *)itm.GetParent())->OnChordList();
}

void HBContextDlg::ScaleList(wxItem &itm, wxCommandEvent& event)
{
  ((HBContextDlg *)itm.GetParent())->OnScaleList();
}


// ---------------------------------------------------------
// Transpose
// ---------------------------------------------------------

class HBTransposer : public wxForm
{
  public:
    HBTransposer(tFilter *f, HBContext *s[], int n);
    virtual void OnOk();
    virtual void OnHelp();

    enum {MAXRES = 8};
    static int analyze_res;

  private:
    tFilter *filter;
    HBContext **seq;
    int n_seq;
    int nn;

    void Iterate(void (HBTransposer::*Action)(tKeyOn *on, tTrack *track, int i));
    void Analyze(tKeyOn *on, tTrack *track, int i);
    void Transpose(tKeyOn *on, tTrack *track, int i);
    void GenerateMapping();

    long count[HBCanvas::SEQMAX * (int)MAXRES][12];
    int  delta[HBCanvas::SEQMAX * (int)MAXRES][12];
    int  MaxCount(int i);
    int  NumCount(int i);
};

int HBTransposer::analyze_res = 1;

HBTransposer::HBTransposer (tFilter * f, HBContext *s[], int n)
	: wxForm( USED_WXFORM_BUTTONS )
{
  filter = f;
  seq = s;
  n_seq = n;
}


void HBTransposer::OnOk()
{
  if (!TrackWin->EventsSelected())
    return;

  nn = n_seq * analyze_res;

  int i, j;
  for (i = 0; i < nn; i++)
  {
    for (j = 0; j < 12; j++)
    {
      count[i][j] = 0;
      delta[i][j] = 0;
    }
  }

  wxBeginBusyCursor();
  filter->Song->NewUndoBuffer();
  Iterate(&HBTransposer::Analyze);
  GenerateMapping();
  Iterate(&HBTransposer::Transpose);
  wxEndBusyCursor();

  wxForm::OnOk();
}

void HBTransposer::OnHelp()
{
        HelpInstance->LoadFile( HelpPathList.FindValidPath("jazz") );
        HelpInstance->KeywordSearch("Harmony browser");
}


int HBTransposer::NumCount(int i)
{
  // count the notes of step i
  int  n = 0;
  for (int k = 0; k < 12; k++)
  {
    if (count[i][k] > 0)
      n++;
  }
  return n;
}

int HBTransposer::MaxCount(int i)
{
  // find the most used note in step i
  int  imax = 0;
  long nmax = count[i][0];
  for (int k = 1; k < 12; k++)
  {
    if (count[i][k] > nmax)
    {
      nmax = count[i][k];
      imax = k;
    }
  }
  return imax;
}

void HBTransposer::GenerateMapping()
{
  int i, j;
  for (i = 0; i < nn; i++)
  {
    HBChord chord = seq[i/analyze_res]->Chord();
    HBChord scale = seq[i/analyze_res]->Scale();
    int notes_remaining = NumCount(i);

    HBChord c;
    int     n;

    // map chord notes to the most used notes

    c = chord;
    n = c.Count();
    for (j = 0; notes_remaining && j < n; j++)
    {
      int old_key = MaxCount(i);
      int new_key = c.Fit(old_key);
      delta[i][old_key] = new_key - old_key;
      c -= new_key;
      count[i][old_key] = 0;
      notes_remaining --;
    }

    // map scale notes to the remaining notes

    c = scale - chord;
    n = c.Count();
    for (j = 0; notes_remaining && j < n; j++)
    {
      int old_key = MaxCount(i);
      int new_key = c.Fit(old_key);
      delta[i][old_key] = new_key - old_key;
      c -= new_key;
      count[i][old_key] = 0;
      notes_remaining --;
    }

    // map remaining notes cromatic

    c = chromatic_scale - scale - chord;
    n = c.Count();
    for (j = 0; notes_remaining && j < n; j++)
    {
      int old_key = MaxCount(i);
      int new_key = c.Fit(old_key);
      delta[i][old_key] = new_key - old_key;
      c -= new_key;
      count[i][old_key] = 0;
      notes_remaining --;
    }

  }
}


void HBTransposer::Analyze(tKeyOn *on, tTrack *track, int i)
{
  count[i][on->Key % 12] += on->Length;
}

void HBTransposer::Transpose(tKeyOn *on, tTrack *track, int i)
{
  tKeyOn *cp = (tKeyOn *)on->Copy();
  cp->Key += delta[i][on->Key % 12];
  track->Kill(on);
  track->Put(cp);
}

void HBTransposer::Iterate(void (HBTransposer::*Action)(tKeyOn *on, tTrack *track, int i))
{

  tTrackIterator Tracks(filter);
  tTrack *t = Tracks.First();
  while (t)
  {
    if (!t->IsDrumTrack())
    {
      tEventIterator Events(t);
      tEvent *e = Events.Range(filter->FromClock, filter->ToClock);
      while (e)
      {
	tKeyOn *on = e->IsKeyOn();
	if (on)
        {
          int i = (int) ((on->Clock - filter->FromClock) * nn / (filter->ToClock - filter->FromClock));
	  (this->*Action)(on, t, i);
	}
	e = Events.Next();
      }
      t->Cleanup();
    }
    t = Tracks.Next();
  }
}


// ------------------- HBFrame ---------------------------


#define MEN_CLOSE	1
#define MEN_MIDI	2
#define MEN_SETTINGS	3
#define MEN_TRANSPOSE	4
#define MEN_CLEARSEQ	6
#define MEN_EDIT	7
#define MEN_MOUSE	8
#define MEN_HELP	9


class HBFrame : public wxFrame, public tHBInterface
{
  public:
    HBFrame(wxFrame *parent);
    ~HBFrame();
    int SeqDefined();
    int GetChordKeys(int *out, int step, int n_steps);
    int GetSelectedChord(int *out);
    int GetSelectedScale(int *out);
    int GetBassKeys(int *out, int step, int n_steps);
  protected:
    virtual void OnMenuCommand(int id);
  private:
    HBCanvas *cnvs;
    int SeqSelected();
};

HBFrame::~HBFrame()
{
  delete cnvs;
}

int HBFrame::SeqDefined()
{
  return cnvs->SeqDefined();
}


int HBFrame::SeqSelected()
{
  if (cnvs->n_seq == 0 || cnvs->mouse_context.SeqNr() == 0)
  {
    wxMessageBox("select a chord from sequence first", "error");
    return 0;
  }
  return 1;
}


int HBFrame::GetChordKeys(int *out, int step, int n_steps)
{
  return cnvs->GetChordKeys(out, step, n_steps);
}

int HBFrame::GetSelectedChord(int *out)
{
  return cnvs->GetSelectedChord(out);
}

int HBFrame::GetSelectedScale(int *out)
{
  return cnvs->GetSelectedScale(out);
}

int HBFrame::GetBassKeys(int *out, int step, int n_steps)
{
  return cnvs->GetBassKeys(out, step, n_steps);
}


void HBFrame::OnMenuCommand(int id)
{
  switch(id)
  {
    case MEN_HELP:
	HelpInstance->LoadFile( HelpPathList.FindValidPath("jazz") );
        HelpInstance->KeywordSearch("Harmony browser");
	break;

    case MEN_MOUSE:
	wxbMessageBox("left: select chord\n"
	              "  +shift: put chord into sequence\n"
	              "middle: same as left+shift\n"
	              "right: play chord\n", "Mousebuttons", wxOK);
	break;

    case MEN_MIDI:
      cnvs->player.SettingsDialog(this);
      break;

    case MEN_CLEARSEQ:
      cnvs->ClearSeq();
      break;
    
    case MEN_SETTINGS:
      cnvs->SettingsDialog(this);
      break;

    case MEN_EDIT:
      {
	if (!SeqSelected())
	  return;
      }
      (void) new HBContextDlg(cnvs, this, cnvs->seq[cnvs->mouse_context.SeqNr() - 1]);
      break;

    case MEN_TRANSPOSE:
      {
	if (!SeqDefined())
	{
          wxMessageBox("define a chord sequence first", "error");
	  return;
	}
	
        wxDialogBox  *panel = new wxDialogBox(this, "Transpose", FALSE);
        HBTransposer *form  = new HBTransposer(TrackWin->Filter, cnvs->seq, cnvs->n_seq);

  	panel->SetLabelPosition(wxHORIZONTAL);
	form->Add(wxMakeFormMessage("Analyze source in sections per harmony"));
	form->Add(wxMakeFormNewLine());
	form->Add(wxMakeFormShort("Sections", &form->analyze_res, wxFORM_DEFAULT, new wxList(wxMakeConstraintRange(1.0, (double)HBTransposer::MAXRES), 0)));
	form->Add(wxMakeFormNewLine());

	form->AssociatePanel(panel);
	panel->Fit();
	panel->Show(TRUE);
      }
      break;

    case MEN_CLOSE:
      the_harmony_browser = 0;
      #if defined(wx_motif) && defined(wxVERSION_NUMBER) && (wxVERSION_NUMBER > 1603)
      wxPostDelete(this);	// Motif doesn't like delete this in menu's???
      #else
      delete this;
      #endif
      break;
  }
}


HBFrame::HBFrame(wxFrame *parent)
  : wxFrame(0, "Harmony Browser", 20, 20, 600, 400)
{
  int w, h;

  wxMenuBar *menu_bar = new wxMenuBar;
  wxMenu    *menu = new wxMenu;
  menu->Append(MEN_SETTINGS,	"&Settings");
  menu->Append(MEN_MIDI,	"&Midi");
  menu->Append(MEN_TRANSPOSE,	"&Transpose");
  menu->Append(MEN_CLEARSEQ,	"&Clear sequence");
  menu->Append(MEN_EDIT,	"&Edit chord");
  menu->Append(MEN_HELP,	"&Help");
  menu->Append(MEN_MOUSE,	"Help &mouse");
  menu->Append(MEN_CLOSE,	"Cl&ose");
  menu_bar->Append(menu,	"&Menu");
  SetMenuBar(menu_bar);

  GetClientSize(&w, &h);
  cnvs =  new HBCanvas(this, 0, 0, w, h);
  //cnvs   =  (HBCanvas *)new dbCanvas(this, 0, 0, w, h);
  //cnvs =  (HBCanvas *)new wxCanvas(this, 0, 0, w, h, 0);
}

void harmony_browser(tEventWin *parent)
{
  if (!the_harmony_browser)
    the_harmony_browser = new HBFrame((wxFrame *)parent);
  ((HBFrame *)the_harmony_browser)->Show(TRUE);
}

