/*********************************************************************
 * locks.c	Warren W. Gay VE3WWG	Sat May 11 15:01:58 1996
 *********************************************************************/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

static int SemUndo = SEM_UNDO;

/*********************************************************************
 * OpenDSPLocks  opens the existing set of 2 semaphores, or creates
 * and  initializes  the new set  if  it  does  not yet exist. This
 * program   chooses   not   to   release  the  IPC  resource  upon
 * termination,  since the  semphore creation and initialization is
 * not  an atomic operation, and can be prone to failure.  However,
 * it  the  IPC resource is removed by some other means (like ipcrm
 * command),  this  function  call  will create a new set using the
 * supplied IPC Key 'LockIPCKey'.
 *
 * NOTES:
 *
 * Two semphores are created in this set. Semaphore # 0 is intended
 * to be used as a /dev/dsp 'input lock' (play lock), and semaphore
 * # 1 is the /dev/dsp 'output lock' (record lock).  I doubt if the
 * record lock  is very useful, but there might be situations where
 * it is needed.
 *********************************************************************/

int OpenDSPLocks(key_t LockIPCKey,int SemUndoFlag) {
	int ipc;
	int s;
	int e;
	int safety = 3;
	union semun u;
	static ushort init_sems[2] = { 1, 1 };
	
	/*************************************************************
	 * Set the SEM_UNDO status :
	 *************************************************************/

	SemUndo = SemUndoFlag ? SEM_UNDO : 0;

	/*************************************************************
	 * If semaphore already exists, just use it :
	 *************************************************************/

	while ( (ipc = semget(LockIPCKey,2,0666)) < 0 && --safety >= 0 ) {	

		/*****************************************************
		 * Failed to find it, try creating it :
		 *****************************************************/

		if ( (ipc = semget(LockIPCKey,2,IPC_CREAT|IPC_EXCL|0666)) < 0
		&&    errno != EEXIST )
			return -1;	/* No system IPC resources? */

		/*****************************************************
		 * Set already exists- timing error? Try again.
		 *****************************************************/

		if ( ipc < 0 ) {
			sleep(1);	/* Allow creator time to init sems */
			continue;	/* Try again */
		}

		/*****************************************************
		 * We created the semaphore set - initialize so that
		 * each semaphore has the value 1 :
		 *****************************************************/

		u.array = &init_sems[0];

		if ( (s = semctl(ipc,0,SETALL,u)) < 0 && errno == EIDRM ) {
			/*********************************************
			 * Another process removed our ipc resource :
			 *********************************************/
			continue;	/* Try again */
		}

		if ( s < 0 ) {
			/*********************************************
			 * We failed to initialize!
			 *********************************************/
			e = errno;			/* Save error */
			semctl(ipc,0,IPC_RMID,NULL);	/* Destroy bad sems */
			errno = e;			/* Restore error */
			return -1;			/* Return err ind. */
		}
	}

	return ipc;	/* Return existing, or newly initialized sems */
}

/*********************************************************************
 * Lock the /dev/dsp device :
 *
 * playrecx :
 *	0	play lock
 *	1	record lock
 *********************************************************************/

int LockDSP(int ipc,int playrecx) {
	int s;
	static struct sembuf sops[1] = { { 0, -1, SEM_UNDO } };

	sops[0].sem_num = playrecx;
	sops[0].sem_flg = SemUndo;

	while ( (s = semop(ipc,sops,1)) < 0 && errno == EINTR ) ;

	return s;
}

/*********************************************************************
 * Lock the /dev/dsp device :
 *
 * playrecx :
 *	0	play lock
 *	1	record lock
 *********************************************************************/

int UnlockDSP(int ipc,int playrecx) {
	int s;
	static struct sembuf sops[1] = { { 0, +1, SEM_UNDO } };

	sops[0].sem_num = playrecx;
	sops[0].sem_flg = SemUndo;

	while ( (s = semop(ipc,sops,1)) < 0 && errno == EINTR ) ;

	return s;
}
