/******************************************************************************/
/* WAV-File player/recorder - Version 0.21                                    */
/* using code of recplay.c from Audio Utilities for Linux 2.0 (c) Liam Corner */
/* (C) FF-Soft '93                                                            */
/* Andre Fuechsel; email: af1@irz.inf.tu-dresden.de /                         */
/*                                                                            */
/* Version 0.21w1 brought to you by:                                          */
/* Warren W. Gay VE3WWG Sun May 12 19:33:30 1996                              */
/******************************************************************************/

#include <stdio.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include "play.h"

extern int errno;
extern char *sys_errlist[];

/*********************************************************************
 * error handler for commandline version on WAVplay/record
 *	severity:	0 = warning -> no exit
 *        	  	1 = error -> exit
 *	text:		error text
 *********************************************************************/

void
usage(const char *cmd,int xc) {

	fprintf(stderr,"Usage: %s [opts] [filename(s)...]\n\n",cmd);
	fputs("Options:\n",stderr);
	fputs("\t-h\tThis info.\n",stderr);
	fputs("\t-q\tQuiet mode (no messages)\n",stderr);
	fputs("\t-s rate\tSet sampling rate\n",stderr);
	fputs("\t-S\tSterio\n",stderr);
	fputs("\t-t secs\tSet time limit\n",stderr);
	fputs("\t-b bits\tSet number of bits (8/12/16)\n",stderr);
	fputs("\t-k key\tSet different IPC Key for lock\n",stderr);
	fputs("\t-r\tRemove and recreate semaphore locks.\n\n",stderr);
	fputs("\t-l\tLock for play.\n",stderr);
	fputs("\t-u\tUnlock for play.\n",stderr);
	fputs("\t-L\tLock for record.\n",stderr);
	fputs("\t-U\tUnlock for record.\n\n",stderr);
	fputs("Normally audio settings are taken from input file,\n"
		"but command options will override them.\n",stderr);
	exit(xc);
}

/*********************************************************************
 * Report an error :
 *********************************************************************/

void err(int syserrflg, char *text) {

	if ( !syserrflg )
		fprintf(stderr,"%s\n",text);
	else	fprintf(stderr,"%s: %s\n",sys_errlist[errno],text);
}

main(int argc, char *argv[]) {
	key_t IPCKey = AUDIOLCK;	/* Default IPC Key for lock */
	int lock = 0;			/* Play lock request */
	int Lock = 0;			/* Record lock request */
	int ipc = -1;			/* Lock IPC ID */
	int n;
	int flgReset = 0;		/* Remove old semaphore set */
	char optch;			/* Option character */
	int mode;			/* PLAY or RECORD */
	header_data data = {0,0,0,0,0,0,0,0};

	if ( (n = strlen(argv[0])) >= 6
	&&   ( !strncmp(argv[0]+n-6,"record",6)
	    || !strncmp(argv[0]+n-6,"wavrec",6) ) )
		mode = RECORD;
	else	mode = PLAY;

	while( (optch = getopt(argc, argv, "qs:St:b:hk:rlLuU")) != -1 )
		switch ( optch ) {
		case 'l' :
		case 'u' :
			lock = optch;
			break;
		case 'L' :
		case 'U' :
			Lock = optch;
			break;
		case 'r':
			flgReset = 1;
			break;
		case 'k':
			IPCKey = (key_t) atol(optarg);
			break;
		case 'q':
			data.quiet_mode = 1;
			break;
		case 's':
			data.speed = atoi(optarg);
			data.force_speed = 1;
			break;
		case 'S':
			data.stereo = data.force_stereo = 1;
			break;
		case 't':
			data.time_limit = atof(optarg);
			break;
		case 'b':
			data.sample_size = atoi(optarg);
			data.force_sample_size = 1;
			break;
		case 'h':
			usage(argv[0],0);
		default:
			usage(argv[0],1);
		}

	/*************************************************************
	 * Obtain semaphores for locking if IPCKey != 0 :
	 *************************************************************/

	if ( IPCKey && (ipc = OpenDSPLocks(IPCKey,!(lock|Lock))) < 0 ) {
		err(1,"Unable to get audio locking semaphores.");
		exit(-1);
	}

	if ( ipc >= 0 && flgReset ) {
		/*****************************************************
		 * Remove old semaphore set :
		 *****************************************************/
		if ( semctl(ipc,0,IPC_RMID,NULL) < 0 ) {
			err(1,"Unable to remove old locking semaphores.");
			exit(-1);
		}

		/*****************************************************
		 * Cancel -u or -U if given :
		 *****************************************************/
		if ( lock == 'u' )
			lock = 'x';	/* Not required now */
		if ( Lock == 'U' )
			lock = 'X';	/* Not required now */

		/*****************************************************
		 * Get replacement semaphores :
		 *****************************************************/
		if ( IPCKey && (ipc = OpenDSPLocks(IPCKey)) < 0 ) {
			err(1,"Unable to get audio locking semaphores.");
			exit(-1);
		}
	}

	/*************************************************************
	 * -u play unlock request :
	 *************************************************************/

	if ( lock == 'u' && UnlockDSP(ipc,0) ) {
		err(1,"-u unlock request failed.");
		exit(1);
	}

	/*************************************************************
	 * -U record unlock request :
	 *************************************************************/

	if ( Lock == 'U' && UnlockDSP(ipc,1) ) {
		err(1,"-U unlock request failed.");
		exit(1);
	}

	/*************************************************************
	 * -l play lock request :
	 *************************************************************/

	if ( lock == 'l' && LockDSP(ipc,0) ) {
		err(1,"-l lock request failed.");
		exit(1);
	}

	/*************************************************************
	 * -L record lock request :
	 *************************************************************/

	if ( Lock == 'L' && LockDSP(ipc,1) ) {
		err(1,"-L lock request failed.");
		exit(1);
	}

	/*************************************************************
	 * If any of -l, -u, -L or -U are given, then we exit(0) here
	 *************************************************************/

	if ( lock || Lock )
		exit(0);	/* Just lock/unlock functions */

	/*************************************************************
	 * Record or Play :
	 *************************************************************/

	if ( mode == RECORD ) {
		/*****************************************************
		 * Set default RECORD options :
		 *****************************************************/
		if ( !data.time_limit )
			data.time_limit = 10.0;	/* 10 secs */
		if ( !data.force_stereo )
			data.stereo = 0;	/* mono */
		if ( !data.force_speed )
			data.speed = 22050;	/* 22050 Hz */
		if ( !data.force_sample_size )
			data.sample_size = 8;	/* 8 bit samples */
	}

	data.lower_border = 0;
	data.upper_border = 0;		/* to be set in recplay! */

	if ( optind >= argc )
		recplay(NULL,mode,data,ipc);
	else	while ( optind < argc ) 
			recplay(argv[optind++],mode,data,ipc);

	exit(0);
}
